#!/usr/bin/env python
# -*- coding: UTF-8 -*-

import click, os, time, fnmatch, json
import re
import sys
import SocketServer
import BaseHTTPServer
import SimpleHTTPServer

import sass, jsmin, coffeescript, BaseHTTPServer 
from watchdog.observers import Observer
from watchdog.events import FileSystemEventHandler
from jsonschema import validate, exceptions

default = {
    "jobs" : [
        {
            "type"   : "javascript",
            "watch"  : ["*.js"]
        },
        {
            "type"   : "sass",
            "watch"  : ["*.scss","*.sass"]
        },
        {
            "type"   : "coffeescript",
            "watch"  : ["*.coffee"]
        }
    ]  
}

job_schema = {
    "type" : "object",
    "properties" : {
        "type" : {"type" : "string"},
        "files" : {"type" : "array"},
        "watch" : {"type" : "array"},
    },
    "required" : ["type","watch"],
}

config_schema = {
    "type" : "object",
    "properties" : {
        "jobs" : { "type" : "array", "items" : job_schema }
    }
}


class Config(object):
    def __init__(self):
        self.config = None

pass_config = click.make_pass_decorator(Config, ensure=True)


class ThreadingSimpleServer(SocketServer.ThreadingMixIn,
                            BaseHTTPServer.HTTPServer):
    pass

 
class checkEvent(FileSystemEventHandler):
    def __init__(self, config, watch_list):
        self.config = config
        self.watch_list = watch_list
    def on_any_event(self, event):
        if event.event_type in ['modified','created']:
            for i in self.watch_list:
                if fnmatch.fnmatch(event.src_path, i) and '.min.' not in event.src_path:
                    click.echo(click.style('Change Detected',bold=True))
                    for j in self.watch_list[i]:
                        do_job(event.src_path, self.config['jobs'][j])


def build_watch_list(jobs):
    watch_list = {}
    for i in jobs:
        for f in i['watch']:
            if f in watch_list:
                if jobs.index(i) not in watch_list[f]:
                    watch_list[f].append( jobs.index(i) )
            else:
                watch_list[f] = [jobs.index(i)]

    return watch_list


def do_job(fpath, job):
    """Simple Build Tool for Webdeveloper"""
    if 'files' not in job:
        files = [fpath] 
    else:
        files = job['files']

    for f in files:
        click.echo(click.style('==>', fg='green')+'%s' % f)
        with open(f) as fo:
            data = fo.read()

        if job['type'] in ['javascript']:
            f = re.sub(r"\.(?=[^.]*$)", r".min.", f) 
            output = jsmin.jsmin(data)
        if job['type'] in ['sass', 'scss']:
            f = re.sub(r"\.scss|\.sass", r".css", f)            
            output = sass.compile(string=data, output_style='compressed')
        if job['type'] in ['coffeescript']:
            f = re.sub(r"\.coffee", r".min.js", f)            
            output = coffeescript.compile(data)
        
        with open(f, 'w') as fh:
            fh.write(data)
        click.echo(click.style('<==', fg='blue')+'%s' % f)


@click.group()
@pass_config
def cli(config):
    '''Jidōka is a build tool for javascript. The name can described as 
    "intelligent automation" or "automation with a human touch."'''

    if os.path.exists('./jidoka.json'):
        with open('./jidoka.json') as f:
            try:
                config.config = json.loads(f.read())
            except ValueError as e:
                click.echo(click.style('Check your Config: %s' % e.message, fg='red'))
                exit()

        try:
            validate(config.config, config_schema)
        except exceptions.ValidationError as e:
            click.echo(click.style('jidoka.json %s' % e.message, fg='red'))
            exit()
    else:
        config.config = default

    pass


@cli.command()
@pass_config
def init(config):
    """Saves a basic project configuration to jidoka.json"""
    if os.path.exists('./jidoka.json'):
        click.echo('This is already a jidoka project!')
    else:
        with open('./jidoka.json','w') as f:
            f.write(json.dumps(config.config, indent=4, sort_keys=True))
        click.echo('Initialise new project, modify jidoka.json to fit your requirements.')


@cli.command()
@pass_config
def serve(config):
    """Start a local HTTPserver"""
    click.echo(click.style('Start development server on http://127.0.0.1:8080', bold=True))
    port = 8000

    server = ThreadingSimpleServer(('', port), SimpleHTTPServer.SimpleHTTPRequestHandler)
    try:
        while True:
            #sys.stdout.flush()
            server.handle_request()
    except KeyboardInterrupt:
        pass


@cli.command()
@pass_config
def build(config):
    """Build your project"""
 
    watch_list = build_watch_list(config.config['jobs'])

    fileList = []
    for root, subFolders, files in os.walk('.'):
        for f in files:
            f = os.path.join(root,f)
            for i in watch_list:
                if fnmatch.fnmatch(f, i) and '.min.' not in f:
                    for j in watch_list[i]:
                        do_job(f, config.config['jobs'][j])


@cli.command()
@pass_config
def watch(config):
    """watch your project and rebuild files"""

    watch_list = build_watch_list(config.config['jobs'])

    event_handler = checkEvent(config.config, watch_list)
    observer = Observer()
    observer.schedule(event_handler, path='.', recursive=True)
    observer.start()

    try:
        while True:
            time.sleep(1)
    except KeyboardInterrupt:
        observer.stop()
    observer.join()


if __name__ == '__main__':
    cli()
